/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "lua/lua.pmh"

namespace fe
{

void SurfaceAccessibleLua::reset(void)
{
	SurfaceAccessibleRecord::reset();
}

BWORD SurfaceAccessibleLua::load(String a_filename,sp<Catalog> a_spSettings)
{
	sp<SurfaceAccessibleI> spDirectInput[4];
	const char* inputName[4]={ "input0", "input1", "input2", "input3" };

	Real frame=0.0;
	String options;
	String inputString;
	if(a_spSettings.isValid())
	{
		frame=a_spSettings->catalog<Real>("frame");
		options=a_spSettings->catalog<String>("options");
		inputString=a_spSettings->catalogOrDefault<String>("inputString","");

		for(I32 m=0;m<4;m++)
		{
			spDirectInput[m]=a_spSettings->catalogOrDefault< sp<Component> >(
					inputName[m],sp<Component>(NULL));
		}
	}

	const String setupString=
			"function createPoint()\n"
			"point = pt:createRecord()\n"
			"output[\"points\"]:add(point)\n"
			"return point\n"
			"end\n"
			"function createPrimitive()\n"
			"primitive = prim:createRecord()\n"
			"output[\"primitives\"]:add(primitive)\n"
			"return primitive\n"
			"end\n"
			"function createVertex(primitive)\n"
			"point = createPoint()\n"
			"pointCount = output[\"points\"]:length()\n"
			"primitive[\"vertexMap\"]:add(pointCount-1)\n"
			"return point\n"
			"end\n\n";

	sp<LuaI> spLuaI=registry()->create("LuaI");

	spLuaI->preloadString(setupString);

	if(a_filename.empty() || !spLuaI->loadFile(a_filename))
	{
		if(inputString.empty() || !spLuaI->loadString(inputString))
		{
			feLog("SurfaceAccessibleLua::load"
					" failed to compile file or string\n");
			reset();
			return FALSE;
		}
	}

	spLuaI->set("frame",frame);

	sp<Scope> spScope=registry()->create("Scope");
	bind(spScope);

	BWORD validInput=FALSE;

	sp<SurfaceAccessibleRecord> spRecordInput[4];
	for(I32 m=0;m<4;m++)
	{
		if(spDirectInput[m].isValid())
		{
			spRecordInput[m]=registry()->create("*.SurfaceAccessibleRecord");
			spRecordInput[m]->bind(spScope);
			spRecordInput[m]->copy(spDirectInput[m]);

			spLuaI->set(inputName[m],spRecordInput[m]->record());

			if(!validInput)
			{
				copy(spRecordInput[m]);
				validInput=TRUE;
			}
		}
	}

	if(!validInput)
	{
		reset();
	}

	spLuaI->set("scope",spScope);
	spLuaI->set("output",record());

	sp<Layout> spLayoutPoint=spScope->declare("pt");
	spLayoutPoint->populate("spc:at","vector3");
	spLuaI->set("pt",spLayoutPoint);

	sp<Layout> spLayoutPrimitive=spScope->declare("prim");
	spLayoutPrimitive->populate("surf:points","intarray");
	spLuaI->set("prim",spLayoutPrimitive);

	SurfaceModel surfaceModelRV;
	surfaceModelRV.bind(record());

	sp<RecordGroup> spRG=surfaceModelRV.surfacePointRG();
	RecordGroup::iterator it;
	if((it=spRG->begin()) != spRG->end())
	{
		spLuaI->set("pointArray",*it);
	}

	spRG=surfaceModelRV.surfaceVertexRG();
	if((it=spRG->begin()) != spRG->end())
	{
		spLuaI->set("vertexArray",*it);
	}

	spRG=surfaceModelRV.surfacePrimitiveRG();
	if((it=spRG->begin()) != spRG->end())
	{
		spLuaI->set("primitiveArray",*it);
	}

	spRG=surfaceModelRV.surfaceDetailRG();
	if((it=spRG->begin()) != spRG->end())
	{
		spLuaI->set("detailArray",*it);
	}

	spLuaI->alias("P","spc:at");
	spLuaI->alias("N","spc:up");
	spLuaI->alias("uv","surf:uvw");
	spLuaI->alias("vertexMap","surf:points");
	spLuaI->alias("points","surf:pointgroup");
	spLuaI->alias("vertices","surf:vertgroup");
	spLuaI->alias("primitives","surf:primgroup");
	spLuaI->alias("detail","surf:detgroup");

	String option;
	while(!(option=options.parse()).empty())
	{
		const String property=option.parse("\"","=");
		const String value=option.parse("\"","=");

		if(value.match("^[-0-9.]+,[-0-9.]+,[-0-9.]+$"))
		{
			SpatialVector vector3(0.0,0.0,0.0);
			sscanf(value.c_str(),"%G,%G,%G",
					&vector3[0],&vector3[1],&vector3[2]);

//			feLog("\"%s\" vector %s\n",property.c_str(),c_print(vector3));
			spLuaI->set(property,vector3);
			continue;
		}
		if(value.match("^[-0-9.]+$"))
		{
			Real number=0.0;
			sscanf(value.c_str(),"%G",&number);

//			feLog("\"%s\" number %.6G\n",property.c_str(),number);
			spLuaI->set(property,number);
			continue;
		}

//		feLog("string \"%s\" \"%s\"\n",property.c_str(),value.c_str());
		spLuaI->set(property,value);
	}

	if(!spLuaI->execute())
	{
		feLog("SurfaceAccessibleLua::load Lua execution failed\n");
		return FALSE;
	}

	return TRUE;
}

} // namespace
