/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SAB_DEBUG		FALSE
#define FE_SAB_COPY_DEBUG	FALSE

#define	FE_SAB_LOCK_CHECK	TRUE
#define	FE_SAB_LOCK_DEBUG	FALSE


#include "surface.pmh"

namespace fe
{

SurfaceAccessibleBase::SurfaceAccessibleBase(void):
	m_paging(FALSE),
	m_locker(-1)
{
}

SurfaceAccessibleBase::~SurfaceAccessibleBase(void)
{
	FEASSERT(m_locker== -1);
}

void SurfaceAccessibleBase::lock(I64 a_id)
{
#if FE_SAB_LOCK_CHECK
#if FE_SAB_LOCK_DEBUG
	feLog("SurfaceAccessibleBase::lock %p vs %p locking \"%s\"\n",
			a_id,m_locker,name().c_str());
#endif

	FEASSERT(a_id!=m_locker);

	if(a_id==m_locker)
	{
		feLog("SurfaceAccessibleBase::lock"
				" refusing to deadlock \"%s\" with %p\n",
				name().c_str(),a_id);
		return;
	}
#endif

	safeLock();

#if FE_SAB_LOCK_CHECK
#if FE_SAB_LOCK_DEBUG
	feLog("SurfaceAccessibleBase::lock %p vs %p locked \"%s\"\n",
			a_id,m_locker,name().c_str());
#endif

	FEASSERT(m_locker== -1);
	m_locker=a_id;
#endif
}

void SurfaceAccessibleBase::unlock(I64 a_id)
{
#if FE_SAB_LOCK_CHECK
#if FE_SAB_LOCK_DEBUG
	feLog("SurfaceAccessibleBase::unlock %p vs %p unlocking \"%s\"\n",
			a_id,m_locker,name().c_str());
#endif

	FEASSERT(m_locker==a_id);

	if(m_locker== -1)
	{
		feLog("SurfaceAccessibleBase::unlock"
				" unexpected unlock of \"%s\" with %p\n",
				name().c_str(),a_id);
		return;
	}
	else if(a_id!=m_locker)
	{
		feLog("SurfaceAccessibleBase::unlock"
				" refusing to unlock \"%s\" with %p when locked with %p\n",
				name().c_str(),a_id,m_locker);
		return;
	}

	m_locker= -1;
#endif

	safeUnlock();

#if FE_SAB_LOCK_CHECK
#if FE_SAB_LOCK_DEBUG
	feLog("SurfaceAccessibleBase::unlock %p vs %p unlocked \"%s\"\n",
			a_id,m_locker,name().c_str());
#endif
#endif
}

//* TODO fix for Maya, or provide Maya version with filtering
void SurfaceAccessibleBase::append(
	sp<SurfaceAccessibleI> a_spSurfaceAccessible,
	const SpatialTransform* a_pTransform,sp<FilterI> a_spFilter)
{
#if FE_SAB_COPY_DEBUG
	feLog("SurfaceAccessibleBase::append\n");
#endif

	sp<SurfaceAccessorI> spInputVertices=
			a_spSurfaceAccessible->accessor(e_primitive,e_vertices);

	sp<SurfaceAccessorI> spOutputVertices=accessor(e_primitive,e_vertices,
			SurfaceAccessibleI::e_createMissing);

	I32 offset[4];

	for(U32 elementPass=0;elementPass<4;elementPass++)
	{
		SurfaceAccessibleI::Element element;

		switch(elementPass)
		{
			case 0:
				element=e_point;
				break;
			case 1:
				element=e_vertex;
				break;
			case 2:
				element=e_primitive;
				break;
			case 3:
				element=e_detail;
				break;
			default:
				;
		}

		sp<SurfaceAccessorI> spOutputGeneric=accessor(element,e_generic);
		offset[elementPass]=
				spOutputGeneric.isValid()? spOutputGeneric->count(): 0;
	}

	if(spInputVertices.isValid() && spOutputVertices.isValid())
	{
		const I32 offsetIndex=offset[0];

		const U32 filterCount=a_spFilter.isValid()?
				a_spFilter->filterCount(): spInputVertices->count();

		I32 pointCount=0;

		//* point indices for each vertex per primitive
		Array< Array<I32> > primVerts;
		primVerts.resize(filterCount);

		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const U32 primitiveIndex=a_spFilter.isValid()?
					a_spFilter->filter(filterIndex): filterIndex;

			Array<I32>& rVertList=primVerts[primitiveIndex];

			const U32 subCount=spInputVertices->subCount(primitiveIndex);

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const I32 pointIndex=
						spInputVertices->integer(primitiveIndex,subIndex);

				rVertList.push_back(offsetIndex+pointIndex);

				if(pointCount<pointIndex+1)
				{
					pointCount=pointIndex+1;
				}
			}
		}

		if(!pointCount)
		{
			//* presume point cloud without primitives

			sp<SurfaceAccessorI> spInputPosition=
					a_spSurfaceAccessible->accessor(e_point,e_position);

			pointCount=
					spInputPosition.isValid()? spInputPosition->count(): 0;
		}

		sp<SurfaceAccessorI> spPointPosition=accessor(e_point,e_position);

#if FE_SAB_COPY_DEBUG
		feLog("SurfaceAccessibleBase::append input pointCount %d\n",
				pointCount);
#endif

		if(spPointPosition.isValid())
		{
#if FALSE
			for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
			{
				spPointPosition->append();
			}
#else
			spPointPosition->append(pointCount);
#endif
		}

		//* WARNING currently, Maya ignores any point append and adds them here
		//* this is different than the Houdini version
		spOutputVertices->append(primVerts);

		if(spOutputVertices->count()!=offset[2]+spInputVertices->count())
		{
			feLog("SurfaceAccessibleBase::append new primitive count %d"
					" does not match prior count %d plus input count %d\n",
					spOutputVertices->count(),offset[2],spInputVertices->count());
			return;
		}
	}

	for(U32 elementPass=0;elementPass<4;elementPass++)
	{
		const I32 offsetIndex=offset[elementPass];

		SurfaceAccessibleI::Element element;

		switch(elementPass)
		{
			case 0:
				element=e_point;
				break;
			case 1:
				element=e_vertex;
				break;
			case 2:
				element=e_primitive;
				break;
			case 3:
				element=e_detail;
				break;
			default:
				;
		}

		Array<Spec> specs;
		a_spSurfaceAccessible->attributeSpecs(specs,element);

		const U32 elementCount=a_spSurfaceAccessible->count(element);
		const U32 specCount=specs.size();

#if FE_SAB_COPY_DEBUG
		feLog("SurfaceAccessibleBase::append"
				" element %d elementCount %d offsetIndex %d specCount %d\n",
				element,elementCount,offsetIndex,specCount);
#endif

		for(U32 specIndex=0;specIndex<specCount;specIndex++)
		{
			const Spec& spec=specs[specIndex];
			const String& specName=spec.name();
			const String& specType=spec.typeName();

			sp<SurfaceAccessorI> spInputAccessor=
					a_spSurfaceAccessible->accessor(element,specName);

			sp<SurfaceAccessorI> spOutputAccessor=accessor(element,specName,
					SurfaceAccessibleI::e_createMissing);

			if(spInputAccessor.isNull() || spOutputAccessor.isNull())
			{
				feLog("SurfaceAccessibleBase::append"
						" \"%s\" \"%s\" accessor failure\n",
						specType.c_str(),specName.c_str());
				continue;
			}

#if FE_SAB_COPY_DEBUG
			feLog("SurfaceAccessibleBase::append %s \"%s\" count %d\n",
					specType.c_str(),specName.c_str(),elementCount);
#endif

			if(specType=="vector3" || specType=="color")
			{
				for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
				{
					SpatialVector vector3=
							spInputAccessor->spatialVector(elementIndex);

					if(a_pTransform)
					{
						if(specName=="P")
						{
							transformVector(*a_pTransform,vector3,vector3);
						}
						else if(specName=="N")
						{
							rotateVector(*a_pTransform,vector3,vector3);
						}
						//* more?
					}

					spOutputAccessor->set(offsetIndex+elementIndex,vector3);
				}
			}
			else if(specType=="real")
			{
				for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
				{
					spOutputAccessor->set(offsetIndex+elementIndex,
							spInputAccessor->real(elementIndex));
				}
			}
			else if(specType=="integer")
			{
				for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
				{
					spOutputAccessor->set(offsetIndex+elementIndex,
							spInputAccessor->integer(elementIndex));
				}
			}
			else if(specType=="string")
			{
				for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
				{
					spOutputAccessor->set(offsetIndex+elementIndex,
							spInputAccessor->string(elementIndex));
				}
			}
		}
#if FE_SAB_COPY_DEBUG
		feLog("SurfaceAccessibleBase::append"
				" output element %d elementCount %d\n",
				element,count(element));
#endif
	}

	sp<SurfaceAccessorI> spInputProperties=a_spSurfaceAccessible->accessor(
			e_primitive,e_properties,e_refuseMissing);

	sp<SurfaceAccessorI> spOutputProperties=accessor(e_primitive,e_properties,
			SurfaceAccessibleI::e_createMissing);

	if(spInputVertices.isValid() && spOutputVertices.isValid() &&
			spInputProperties.isValid() && spOutputProperties.isValid())
	{
		const I32 offsetIndex=offset[2];

		const U32 filterCount=a_spFilter.isValid()?
				a_spFilter->filterCount(): spInputVertices->count();

		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const U32 primitiveIndex=a_spFilter.isValid()?
					a_spFilter->filter(filterIndex): filterIndex;

			const BWORD openCurve=spInputProperties->integer(primitiveIndex,
					SurfaceAccessibleI::e_openCurve);
			spOutputProperties->set(offsetIndex+filterIndex,
					SurfaceAccessibleI::e_openCurve,openCurve);
		}
	}

#if FE_SAB_COPY_DEBUG
		feLog("SurfaceAccessibleBase::append done\n");
#endif
}

void SurfaceAccessibleBase::instance(
	sp<SurfaceAccessibleI> a_spSurfaceAccessible,
	const Array<SpatialTransform>& a_rTransformArray)
{
	const U32 transformCount=a_rTransformArray.size();
	for(U32 transformIndex=0;transformIndex<transformCount;transformIndex++)
	{
		append(a_spSurfaceAccessible,&a_rTransformArray[transformIndex]);
	}
}

I32 SurfaceAccessibleBase::count(SurfaceAccessibleI::Element a_element) const
{
//	feLog("SurfaceAccessibleBase::count element %d\n",a_element);

	switch(a_element)
	{
		case e_point:
		{
			sp<SurfaceAccessorI> spAccessor=
					const_cast<SurfaceAccessibleBase*>(this)->accessor(
					e_point,e_position);

			if(spAccessor.isNull())
			{
				feLog("SurfaceAccessibleBase::count point accessor null\n");
				return 0;
			}

			return spAccessor->count();
		}
		case e_primitive:
		{
			sp<SurfaceAccessorI> spAccessor=
					const_cast<SurfaceAccessibleBase*>(this)->accessor(
					e_primitive,e_vertices);

			if(spAccessor.isNull())
			{
//				feLog("SurfaceAccessibleBase::count primitive accessor null\n");
				return 0;
			}

			return spAccessor->count();
		}
		default:
			;
	}

	return 0;
}

BWORD SurfaceAccessibleBase::discard(SurfaceAccessibleI::Element a_element,
	SurfaceAccessibleI::Attribute a_attribute)
{
	String name;
	switch(a_attribute)
	{
		case SurfaceAccessibleI::e_position:
			name="P";
			break;
		case SurfaceAccessibleI::e_normal:
			name="N";
			break;
		case SurfaceAccessibleI::e_uv:
			name="uv";
			break;
		default:
			return FALSE;
	}

	return discard(a_element,name);
}

sp<SurfaceI> SurfaceAccessibleBase::surface(String a_group)
{
//	feLog("SurfaceAccessibleBase::surface(\"%s\")\n",a_group.c_str());

	//* TODO select form
	//* TODO group
	sp<SurfaceI> spSurface;

	FEASSERT(registry().isValid());

	if(registry().isValid())
	{
		sp<SurfaceAccessorI> spSurfacePoints=
				accessor(e_point,e_position,e_refuseMissing);
		const I32 pointCount=spSurfacePoints.isValid()?
				spSurfacePoints->count(): 0;

		sp<SurfaceAccessorI> spSurfaceVertices=
				accessor(e_primitive,e_vertices,e_refuseMissing);
		const I32 primitiveCount=spSurfaceVertices.isValid()?
				spSurfaceVertices->count(): 0;

		const BWORD pointCloud=(pointCount && !primitiveCount);

		sp<SurfaceAccessorI> spSurfaceProperties=
				accessor(e_primitive,e_properties,e_refuseMissing);

		const BWORD openCurve=(spSurfaceProperties.isValid() &&
				spSurfaceProperties->count() &&
				spSurfaceProperties->integer(0,
				SurfaceAccessibleI::e_openCurve));

		if(pointCloud || openCurve)
		{
			spSurface=registry()->create(
					"SurfaceI.*.*.SurfaceCurvesAccessible");

			sp<SurfaceCurvesAccessible> spSurfaceCurvesAccessible=
					spSurface;
			if(spSurfaceCurvesAccessible.isValid())
			{
				spSurfaceCurvesAccessible->bind(
						sp<SurfaceAccessibleI>(this));
			}
		}
		else
		{
			spSurface=registry()->create(
					"SurfaceI.*.*.SurfaceTrianglesAccessible");

			sp<SurfaceTrianglesAccessible> spSurfaceTrianglesAccessible=
					spSurface;
			if(spSurfaceTrianglesAccessible.isValid())
			{
				spSurfaceTrianglesAccessible->bind(
						sp<SurfaceAccessibleI>(this));
			}
		}
	}
	else
	{
		feLog("SurfaceAccessibleBase::surface registry is NULL\n");
	}

	if(spSurface.isNull())
	{
		feLog("SurfaceAccessibleBase::surface failed to access surface\n");
	}

	return spSurface;
}

sp<SurfaceI> SurfaceAccessibleBase::subSurface(U32 a_subIndex,String a_group)
{
	//* TODO

	return sp<SurfaceI>(NULL);
}

sp<SpannedRange> SurfaceAccessibleBase::atomize(
	AtomicChange a_atomicChange,String a_group,U32 a_desiredCount)
{
#if FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::atomize\n");
#endif

	if(a_atomicChange==e_pointsOfPrimitives)
	{
		const BWORD checkPages=TRUE;
		return atomizeConnectivity(a_atomicChange,a_group,
				a_desiredCount,checkPages);
	}

	//* fallback to non-atomic

	const BWORD asPoints=(a_atomicChange==e_pointsOnly);
	const Element element=asPoints? e_point: e_primitive;
	sp<SurfaceAccessorI> spAccessor=accessor(element,e_generic);
	const I32 elementCount=spAccessor.isValid()? spAccessor->count(): 0;
	sp<SpannedRange> spSpannedRange(new SpannedRange());

	//* use everything

	if(a_group.empty())
	{
		if(elementCount>0)
		{
			spSpannedRange->nonAtomic().add(0,elementCount-1);
		}
		return spSpannedRange;
	}

	//* use group

//	feLog("SurfaceAccessibleBase::atomize group \"%s\"\n",a_group.c_str());

	SpannedRange::MultiSpan& rMultiSpan=spSpannedRange->nonAtomic();

#if FALSE
	I32 spanStart= -1;
	for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
	{
		const I32 elementPage=
				asPoints? pointPage(elementIndex): primitivePage(elementIndex);
		if(spanStart<0 && elementPage>=0)
		{
			//* begin span
			spanStart=elementIndex;
		}
		else if(spanStart>=0 && elementPage<0)
		{
			//* complete span
			rMultiSpan.add(spanStart,elementIndex-1);
			spanStart=-1;
		}
	}

	//* complete last span
	if(spanStart>=0)
	{
		rMultiSpan.add(spanStart,elementCount-1);

	}
#else
	const Element groupElement=asPoints? e_pointGroup: e_primitiveGroup;

	sp<SurfaceAccessorI> spElementGroup=accessor(groupElement,a_group);

	spElementGroup->appendGroupSpans(rMultiSpan);
#endif

	return spSpannedRange;
}

sp<SpannedRange> SurfaceAccessibleBase::atomizeConnectivity(
	AtomicChange a_atomicChange,String a_group,U32 a_desiredCount,
	BWORD a_checkPages)
{
	if(a_checkPages)
	{
		preparePaging(a_atomicChange,a_group);
	}

	sp<SpannedRange> spSpannedRange(new SpannedRange());

	sp<SurfaceAccessorI> spPoints=accessor(e_point,e_position);
	sp<SurfaceAccessorI> spVertices=accessor(e_primitive,e_vertices);

	const U32 pointCount=spPoints->count();
	const U32 primitiveCount=spVertices->count();

	Array<U32> pointPatch(pointCount,0);
	Array<I32> primitivePatch(primitiveCount,0);

	Array<U32> pagePatch(primitiveCount+1,0);	//* pages <= prims+1
	Array<U32> histogram(primitiveCount+1,0);	//* patches <= prims+1

	U32 maxPatch=0;
	U32 maxPage=0;

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		if(a_checkPages && primitivePage(primitiveIndex)<0)
		{
			FEASSERT(primitiveIndex<primitiveCount);
			primitivePatch[primitiveIndex]= -1;
			continue;
		}

		U32 lowPatch=0;
		U32 lowPage=0;
		BWORD multiPatch=FALSE;
		BWORD zeroPatches=FALSE;
		BWORD multiPage=FALSE;

		const U32 subCount=spVertices->subCount(primitiveIndex);
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const U32 pointIndex=spVertices->integer(primitiveIndex,subIndex);
			FEASSERT(pointIndex<pointCount);
			U32& rPatch=pointPatch[pointIndex];
			FEASSERT(rPatch<=maxPatch);

			const I32 page=a_checkPages? pointPage(pointIndex): 0;
			if(page>0)
			{
				FEASSERT(page<=I32(primitiveCount));
				const U32 existingPatch=pagePatch[page];
				FEASSERT(existingPatch<=maxPatch);

				if(existingPatch)
				{
					rPatch=existingPatch;
				}
				if(!lowPage)
				{
					lowPage=page;
				}
				else if(I32(lowPage)!=page)
				{
					multiPage=TRUE;
					if(I32(lowPage)>page)
					{
						lowPage=page;
					}
				}
			}
			if(rPatch)
			{
				if(!lowPatch)
				{
					lowPatch=rPatch;
				}
				else if(lowPatch!=rPatch)
				{
					multiPatch=TRUE;
					if(lowPatch>rPatch)
					{
						lowPatch=rPatch;
					}
				}
			}
			else
			{
				zeroPatches=TRUE;
			}
		}

		if(multiPage)
		{
			//* leave for postAtomic
			lowPatch=0;
		}
		else if(multiPatch)
		{
//			feLog("multiPatch\n");

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				const U32 patch=pointPatch[pointIndex];

//				feLog("prim %d sub %d/%d patch %d\n",
//						primitiveIndex,subIndex,subCount,patch);

				if(patch<1)
				{
//					feLog("assign patch %d to lowPatch %d\n",patch,lowPatch);

					pointPatch[pointIndex]=lowPatch;
				}
				else if(patch!=lowPatch)
				{
//					feLog("connected patch %d != lowPatch %d\n",patch,lowPatch);

					for(U32 pointOther=0;pointOther<pointCount;pointOther++)
					{
						FEASSERT(pointOther<pointCount);
						U32& rOtherPatch=pointPatch[pointOther];

						if(rOtherPatch==patch)
						{
//							feLog("%d merge patch %d into %d\n",
//									pointIndex,rOtherPatch,lowPatch);

							//* merge all 'patch' into 'lowPatch'
							rOtherPatch=lowPatch;
						}
						else if(rOtherPatch==maxPatch)
						{
//							feLog("%d shift patch %d into %d\n",
//									pointIndex,rOtherPatch,patch);

							//* shift all 'maxPatch' into emptied 'patch'
							rOtherPatch=patch;
						}
					}

					for(U32 primitiveOther=0;primitiveOther<primitiveIndex;
							primitiveOther++)
					{
						I32& rOtherPrim=primitivePatch[primitiveOther];
						primitivePatch[primitiveIndex]=lowPatch;
						if(rOtherPrim==I32(patch))
						{
//							feLog("%d merge patch %d into %d\n",
//									primitiveOther,rOtherPrim,lowPatch);

							//* merge all 'patch' into 'lowPatch'
							rOtherPrim=lowPatch;
						}
						else if(rOtherPrim==I32(maxPatch))
						{
//							feLog("%d shift prim %d into %d\n",
//									primitiveOther,rOtherPrim,patch);

							//* shift all 'maxPatch' into emptied 'patch'
							rOtherPrim=patch;
						}
					}

					histogram[lowPatch]+=histogram[patch];
					histogram[patch]=histogram[maxPatch];

					FEASSERT(maxPatch);
					maxPatch--;

//					feLog("maxPatch -> %d\n",maxPatch);
				}
			}
		}
		else if(!lowPatch)
		{
			maxPatch++;
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				pointPatch[pointIndex]=maxPatch;
			}
			lowPatch=maxPatch;
		}
		else if(zeroPatches)
		{
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				const U32 pointIndex=
						spVertices->integer(primitiveIndex,subIndex);
				FEASSERT(pointIndex<pointCount);
				pointPatch[pointIndex]=lowPatch;
			}
		}

//		feLog("primitive %d/%d lowPatch %d/%d\n",
//				primitiveIndex,primitiveCount,lowPatch,maxPatch);

		FEASSERT(lowPatch<=maxPatch);

		FEASSERT(primitiveIndex<primitiveCount);
		primitivePatch[primitiveIndex]=lowPatch;

		FEASSERT(lowPatch<=primitiveCount);
		histogram[lowPatch]+=subCount;

		if(lowPatch)
		{
			FEASSERT(lowPage<=primitiveCount);
			pagePatch[lowPage]=lowPatch;
		}
		if(maxPage<lowPage)
		{
			maxPage=lowPage;
		}
	}

#if FE_SAB_DEBUG
	feLog("SurfaceAccessibleBase::atomizeConnectivity maxPage %d\n",maxPage);
	for(U32 pageIndex=0;pageIndex<=maxPage;pageIndex++)
	{
		FEASSERT(pageIndex<=primitiveCount);
		feLog("page %d uses patch %d\n",pageIndex,pagePatch[pageIndex]);
	}
	feLog("SurfaceAccessibleBase::atomizeConnectivity maxPatch %d\n",maxPatch);
	for(U32 patchIndex=0;patchIndex<=maxPatch;patchIndex++)
	{
		FEASSERT(patchIndex<=primitiveCount);
		feLog("patch %d has %d points\n",patchIndex,histogram[patchIndex]);
	}
#endif

	const U32 desiredSize=a_desiredCount? primitiveCount/a_desiredCount: 1;

	Array<U32> patchAtom(maxPatch+1,0);

	I32 maxAtom=1;
	U32 atomSize=0;
	for(U32 patchIndex=1;patchIndex<=maxPatch;patchIndex++)
	{
		if(atomSize>=desiredSize)
		{
			maxAtom++;
			atomSize=0;
		}
		FEASSERT(patchIndex<=maxPatch);
		patchAtom[patchIndex]=maxAtom;

		FEASSERT(patchIndex<=primitiveCount);
		atomSize+=histogram[patchIndex];
#if FE_SAB_DEBUG
		feLog("patch %d atom %d\n",patchIndex,maxAtom);
#endif
	}

	Array<SpannedRange::MultiSpan*> multiSpans(maxAtom+1,NULL);
	I32 spanAtom= -1;
	I32 spanStart= -1;
#if FE_SAB_DEBUG
	feLog("first %d from %d\n",spanAtom,spanStart);
#endif
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		FEASSERT(primitiveIndex<primitiveCount);
		const I32 patchIndex=primitivePatch[primitiveIndex];

		FEASSERT(patchIndex<0 || patchIndex<=I32(maxPatch));
		const I32 atomIndex=patchIndex>=0? patchAtom[patchIndex]: -1;

		FEASSERT(atomIndex<=maxAtom);
		FEASSERT(spanAtom<=maxAtom);
		if(atomIndex!=spanAtom)
		{
			//* complete existing span
			SpannedRange::MultiSpan* pMultiSpan=NULL;
			if(spanAtom>=0)
			{
				FEASSERT(spanAtom<=maxAtom);
				pMultiSpan=multiSpans[spanAtom];
				if(!pMultiSpan)
				{
					pMultiSpan=spanAtom? &spSpannedRange->addAtomic():
							&spSpannedRange->postAtomic();
					multiSpans[spanAtom]=pMultiSpan;
				}
			}
			if(spanStart>=0)
			{
#if FE_SAB_DEBUG
				feLog("complete %d span %d %d\n",
						spanAtom,spanStart,primitiveIndex-1);
#endif
				pMultiSpan->add(spanStart,primitiveIndex-1);
				spanStart= -1;
			}

			//* start another span
			spanAtom=atomIndex;
			if(atomIndex>=0)
			{
				spanStart=primitiveIndex;

#if FE_SAB_DEBUG
				feLog("start %d from %d\n",spanAtom,spanStart);
#endif
			}
		}
	}

	//* complete last span
	SpannedRange::MultiSpan* pMultiSpan=NULL;
	if(spanStart>=0)
	{
		FEASSERT(spanAtom<=maxAtom);
		pMultiSpan=multiSpans[spanAtom];
		if(!pMultiSpan)
		{
			pMultiSpan=spanAtom? &spSpannedRange->addAtomic():
					&spSpannedRange->postAtomic();
			multiSpans[spanAtom]=pMultiSpan;
		}
		pMultiSpan->add(spanStart,primitiveCount-1);
	}

#if FE_SAB_DEBUG
	feLog("brief: %s\n",spSpannedRange->brief().c_str());
	feLog("dump:\n%s\n",spSpannedRange->dump().c_str());
#endif

	return spSpannedRange;
}

//* static
String SurfaceAccessibleBase::elementLayout(
	SurfaceAccessibleI::Element a_element)
{
	switch(a_element)
	{
		case e_point:
			return "pt";
		case e_vertex:
			return "vtx";
		case e_primitive:
			return "prim";
		case e_detail:
			return "detail";
		default:
			;
	}

	return "";
}

//* static
String SurfaceAccessibleBase::commonName(String a_attrName)
{
	if(a_attrName=="spc:at")
	{
		return "P";
	}
	else if(a_attrName=="spc:up")
	{
		return "N";
	}
	else if(a_attrName=="surf:uvw")
	{
		return "uv";
	}
	else if(a_attrName=="surf:color")
	{
		return "Cd";
	}

	return a_attrName;
}

//* static
String SurfaceAccessibleBase::attributeName(String a_specName)
{
	if(a_specName=="P")
	{
		return "spc:at";
	}
	else if(a_specName=="N")
	{
		return "spc:up";
	}
	else if(a_specName=="uv")
	{
		return "surf:uvw";
	}
	else if(a_specName=="Cd")
	{
		return "surf:color";
	}

	return a_specName;
}

void SurfaceAccessibleBase::outlineClear(void)
{
	m_outline.clear();
}

void SurfaceAccessibleBase::outlineAppend(String a_line)
{
	const I32 outlineCount=m_outline.size();
	m_outline.resize(outlineCount+1);
	m_outline[outlineCount]=a_line;
}

BWORD SurfaceAccessibleBase::copyOutline(Array<String>& a_rStringArray) const
{
	const I32 outlineCount=m_outline.size();
	a_rStringArray.resize(outlineCount);

	for(I32 outlineIndex=0;outlineIndex<outlineCount;outlineIndex++)
	{
		a_rStringArray[outlineIndex]=m_outline[outlineIndex];
	}

	return TRUE;
}

void SurfaceAccessibleBase::outlineCreateDefault(void)
{
	const String nameAttr=findNameAttribute();

	sp<SurfaceAccessorI> spPrimitiveName=accessor(e_primitive,nameAttr);
	if(spPrimitiveName.isValid())
	{
		std::set<String> nameSet;

		const I32 primitiveCount=spPrimitiveName->count();
		for(I32 primitiveIndex=0;primitiveIndex<primitiveCount;
				primitiveIndex++)
		{
			const String name=spPrimitiveName->string(primitiveIndex);
			if(nameSet.find(name)==nameSet.end())
			{
				nameSet.insert(name);
			}
		}

		for(std::set<String>::iterator it=nameSet.begin();
				it!=nameSet.end();it++)
		{
			const String name= *it;
			outlineAppend(name);
		}
	}
}

String SurfaceAccessibleBase::findNameAttribute(void)
{
	String partAttr;

	Array<SurfaceAccessibleI::Spec> specs;
	attributeSpecs(specs,SurfaceAccessibleI::e_primitive);

	const I32 specCount=specs.size();
	for(I32 specIndex=0;specIndex<specCount;specIndex++)
	{
		const SurfaceAccessibleI::Spec& rSpec=specs[specIndex];

		const String attrName=rSpec.name();
		if(attrName!="name" && attrName!="part")
		{
			continue;
		}

		const String attrType=rSpec.typeName();
		if(attrType!="string")
		{
			continue;
		}

		String valueString;

		sp<SurfaceAccessorI> spSurfaceAccessorI=accessor(e_primitive,attrName,
				SurfaceAccessibleI::e_refuseMissing);

		if(spSurfaceAccessorI.isValid())
		{
			partAttr=attrName;

			//* priority
			if(attrName=="name")
			{
				break;
			}
		}
	}

	return partAttr;
}


} // namespace
